#include <cmath>
#include <QDateTime>
#include <QSet>
#include "kblight.h"
#include "kbmode.h"

static int _shareDimming = -1;
static QSet<KbLight*> activeLights;

KbLight::KbLight(KbMode* parent, const KeyMap& keyMap) :
    QObject(parent), _previewAnim(0), lastFrameSignal(0), _dimming(0), _start(false), _needsSave(true)
{
    map(keyMap);
}

KbLight::KbLight(KbMode* parent, const KeyMap& keyMap, const KbLight& other) :
    QObject(parent), _previewAnim(0), _map(other._map), _colorMap(other._colorMap), lastFrameSignal(0), _dimming(other._dimming), _start(false), _needsSave(true)
{
    map(keyMap);
    // Duplicate animations
    foreach(KbAnim* animation, other._animList)
        _animList.append(new KbAnim(this, keyMap, *animation));
}

void KbLight::map(const KeyMap& map){
    // If any of the keys are missing from the color map, set them to white
    QHashIterator<QString, Key> i(map);
    while(i.hasNext()){
        i.next();
        const QString& key = i.key();
        if(!_colorMap.contains(key))
            _colorMap[key] = 0xFFFFFFFF;
    }
    // Set the new map
    _map = map;
    foreach(KbAnim* anim, _animList)
        anim->map(map);
    _needsSave = true;
    emit updated();
}

KbLight::~KbLight(){
    activeLights.remove(this);
}

void KbLight::color(const QColor& newColor){
    QRgb newRgb = newColor.rgb();
    QMutableHashIterator<QString, QRgb> i(_colorMap);
    while(i.hasNext()){
        i.next();
        i.value() = newRgb;
    }
    _needsSave = true;
}

int KbLight::shareDimming(){
    return _shareDimming;
}

void KbLight::shareDimming(int newShareDimming){
    if(_shareDimming == newShareDimming)
        return;
    _shareDimming = newShareDimming;
    if(newShareDimming != -1){
        foreach(KbLight* light, activeLights)
            light->dimming(newShareDimming);
    }
}

void KbLight::dimming(int newDimming){
    if(_shareDimming != -1)
        shareDimming(newDimming);
    _needsSave = true;
    _dimming = newDimming;
    emit updated();
}

KbAnim* KbLight::addAnim(const AnimScript *base, const QStringList &keys, const QString& name, const QMap<QString, QVariant>& preset){
    // Stop and restart all existing animations
    stopPreview();
    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
    foreach(KbAnim* anim, _animList){
        anim->stop();
        anim->trigger(timestamp);
    }
    // Load the new animation and set preset parameters
    KbAnim* anim = new KbAnim(this, _map, name, keys, base);
    QMapIterator<QString, QVariant> i(preset);
    while(i.hasNext()){
        i.next();
        anim->parameter(i.key(), i.value());
    }
    anim->commitParams();
    // Add the animation and start it
    _animList.append(anim);
    anim->trigger(timestamp);
    _start = true;
    _needsSave = true;
    return anim;
}

void KbLight::previewAnim(const AnimScript* base, const QStringList& keys, const QMap<QString, QVariant>& preset){
    if(_previewAnim)
        stopPreview();
    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
    // Load the new animation and set preset parameters
    KbAnim* anim = new KbAnim(this, _map, "", keys, base);
    QMapIterator<QString, QVariant> i(preset);
    while(i.hasNext()){
        i.next();
        anim->parameter(i.key(), i.value());
    }
    anim->commitParams();
    anim->reInit();
    // Add the animation and start it
    _previewAnim = anim;
    anim->trigger(timestamp);
    _start = true;
}

void KbLight::stopPreview(){
    delete _previewAnim;
    _previewAnim = 0;
}

KbAnim* KbLight::duplicateAnim(KbAnim* oldAnim){
    // Stop and restart all existing animations
    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
    foreach(KbAnim* anim, _animList){
        anim->stop();
        anim->trigger(timestamp);
    }
    // Same as addAnim, just duplicate the existing one
    KbAnim* anim = new KbAnim(this, _map, *oldAnim);
    anim->newId();
    int index = _animList.indexOf(oldAnim);
    if(index < 0)
        _animList.append(anim);
    else
        _animList.insert(index + 1, anim);
    anim->trigger(timestamp);
    _start = true;
    _needsSave = true;
    return anim;
}

bool KbLight::isStarted(){
    if(!_start)
        return false;
    foreach(KbAnim* animation, _animList){
        if(!animation->isRunning())
            return false;
    }
    return true;
}

void KbLight::restartAnimation(){
    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
    foreach(KbAnim* anim, _animList){
        anim->stop();
        anim->trigger(timestamp);
    }
    stopPreview();
    _start = true;
}

void KbLight::animKeypress(const QString& key, bool down){
    foreach(KbAnim* anim, _animList){
        if(anim->keys().contains(key))
            anim->keypress(key, down, QDateTime::currentMSecsSinceEpoch());
    }
    if(_previewAnim){
        if(_previewAnim->keys().contains(key))
            _previewAnim->keypress(key, down, QDateTime::currentMSecsSinceEpoch());
    }
}

void KbLight::open(){
    // Apply shared dimming if needed
    if(_shareDimming != -1 && _shareDimming != _dimming)
        dimming(_shareDimming);
    activeLights.insert(this);
    if(_start)
        return;
    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
    foreach(KbAnim* anim, _animList)
        anim->trigger(timestamp);
    if(_previewAnim)
        _previewAnim->trigger(timestamp);
    _start = true;
}

void KbLight::close(){
    activeLights.remove(this);
    foreach(KbAnim* anim, _animList)
        anim->stop();
    stopPreview();
    _start = false;
}

void KbLight::printRGB(QFile& cmd, const QHash<QString, QRgb>& animMap){
    QHashIterator<QString, QRgb> i(animMap);
    while(i.hasNext()){
        i.next();
        QString name = i.key();
        QRgb color = i.value();
        // Make sure the key is in the map before printing it
        const Key& key = _map[name];
        if(!key.hasLed)
            continue;
        cmd.write(" ");
        cmd.write(name.toLatin1());
        char output[8];
        snprintf(output, sizeof(output), ":%02x%02x%02x", qRed(color), qGreen(color), qBlue(color));
        cmd.write(output);
    }
}

// Colorspace conversion: linear <-> sRGB
// (sRGB: [0, 255], linear: [0, 1])

static float sToL(float srgb){
    srgb /= 255.f;
    if(srgb <= 0.04045f)
        return srgb / 12.92f;
    return pow((srgb + 0.055f) / 1.055f, 2.4f);
}

static float lToS(float linear){
    if(linear <= 0.0031308f)
        return 12.92f * linear * 255.f;
    return (1.055f * pow(linear, 1.f / 2.4f) - 0.055f) * 255.f;
}

// Convert RGB to monochrome
QRgb monoRgb(float r, float g, float b){
    // It's important to use a linear colorspace for this, otherwise the colors will appear inconsistent
    // Note that although we could use linear space for alpha blending or the animation blending functions, we don't.
    // The reason for this is that photo manipulation programs don't do it either, so even though the result would technically be more correct,
    // it would look wrong to most people.
    r = sToL(r);
    g = sToL(g);
    b = sToL(b);
    int value = round(lToS((r + g + b) / 3.f));
    return qRgb(value, value, value);
}

void KbLight::frameUpdate(QFile& cmd, const ColorMap& indicators, bool monochrome){
    // Advance animations
    ColorMap animMap = _colorMap;
    quint64 timestamp = QDateTime::currentMSecsSinceEpoch();
    foreach(KbAnim* anim, _animList)
        anim->blend(animMap, timestamp);
    if(_previewAnim)
        _previewAnim->blend(animMap, timestamp);

    // Set brightness and indicators
    QStringList activeIndicators;
    if(monochrome || !indicators.isEmpty()){
        QMutableHashIterator<QString, QRgb> i(animMap);
        while(i.hasNext()){
            i.next();
            const QString& key = i.key();
            QRgb& rgb = i.value();
            float r = qRed(rgb);
            float g = qGreen(rgb);
            float b = qBlue(rgb);
            // Apply indicators
            if(indicators.contains(key)){
                const QRgb& rgb2 = indicators.value(key);
                float r2 = qRed(rgb2);
                float g2 = qGreen(rgb2);
                float b2 = qBlue(rgb2);
                float a2 = qAlpha(rgb2) / 255.f;
                float a = 1.f;
                float a3 = a2 + (1.f - a2) * a;
                r = round((r2 * a2 + r * a * (1.f - a2)) / a3);
                g = round((g2 * a2 + g * a * (1.f - a2)) / a3);
                b = round((b2 * a2 + b * a * (1.f - a2)) / a3);
                activeIndicators << key;
            }
            // If monochrome mode is active, average them all out to get a grayscale image
            if(monochrome)
                rgb = monoRgb(r, g, b);
            else
                rgb = qRgb(r, g, b);
        }
    }

    // Emit signals for the animation (only do this every 50ms - it can cause a lot of CPU usage)
    if(timestamp >= lastFrameSignal + 50){
        emit frameDisplayed(animMap, activeIndicators);
        lastFrameSignal = timestamp;
    }

    cmd.write(QString().sprintf("rgb").toLatin1());
    // If brightness is at 0%, turn off lighting entirely
    if(_dimming == 3){
        cmd.write(" 000000");
        return;
    }

    float light = (3 - _dimming) / 3.f;
    // Apply global dimming
    if(light != 1.f){
        QMutableHashIterator<QString, QRgb> i(animMap);
        while(i.hasNext()){
            i.next();
            QRgb& rgb = i.value();
            // Like the monochrome conversion, we want a linear colorspace for this
            float r = sToL(qRed(rgb));
            float g = sToL(qGreen(rgb));
            float b = sToL(qBlue(rgb));
            r *= light;
            g *= light;
            b *= light;
            rgb = qRgb(round(lToS(r)), round(lToS(g)), round(lToS(b)));
        }
    }

    // Apply light
    printRGB(cmd, animMap);
}

void KbLight::base(QFile &cmd, bool ignoreDim, bool monochrome){
    close();
    if(_dimming == MAX_DIM && !ignoreDim){
        cmd.write(QString().sprintf("rgb 000000").toLatin1());
        return;
    }
    // Set just the background color, ignoring any animation
    cmd.write(QString().sprintf("rgb").toLatin1());
    QHash<QString, QRgb> animMap = _colorMap;
    // If monochrome is active, create grayscale
    if(monochrome){
        QMutableHashIterator<QString, QRgb> i(animMap);
        while(i.hasNext()){
            i.next();
            QRgb& rgb = i.value();
            rgb = monoRgb(qRed(rgb), qGreen(rgb), qBlue(rgb));
        }
    }
    animMap["mr"] = qRgb(0, 0, 0);
    animMap["m1"] = qRgb(0, 0, 0);
    animMap["m2"] = qRgb(0, 0, 0);
    animMap["m3"] = qRgb(0, 0, 0);
    animMap["lock"] = qRgb(0, 0, 0);
    printRGB(cmd, animMap);
}

void KbLight::load(CkbSettings& settings){
    // Load light settings
    _needsSave = false;
    SGroup group(settings, "Lighting");
    KeyMap currentMap = _map;
    _map = KeyMap::fromName(settings.value("KeyMap").toString());
    _dimming = settings.value("Brightness").toUInt();
    if(_dimming > MAX_DIM)
        _dimming = MAX_DIM;
    // Load RGB settings
    bool useReal = settings.value("UseRealNames").toBool();
    {
        SGroup group(settings, "Keys");
        foreach(QString key, settings.childKeys()){
            QString name = key.toLower();
            if(!useReal)
                name = _map.fromStorage(name);
            QColor color = settings.value(key).toString();
            if(!color.isValid())
                color = QColor(255, 255, 255);
            _colorMap[name] = color.rgb();
        }
    }
    // Load animations
    foreach(KbAnim* anim, _animList)
        anim->deleteLater();
    _animList.clear();
    {
        SGroup group(settings, "Animations");
        foreach(QString anim, settings.value("List").toStringList()){
            QUuid id = anim;
            _animList.append(new KbAnim(this, _map, id, settings));
        }
    }
    emit didLoad();
    map(currentMap);
}

void KbLight::save(CkbSettings& settings){
    _needsSave = false;
    SGroup group(settings, "Lighting");
    settings.setValue("KeyMap", _map.name());
    settings.setValue("Brightness", _dimming);
    settings.setValue("UseRealNames", true);
    {
        // Save RGB settings
        SGroup group(settings, "Keys");
        foreach(QString key, _colorMap.keys())
            settings.setValue(key, QColor(_colorMap.value(key)).name());
    }
    {
        // Save animations
        SGroup group(settings, "Animations");
        QStringList aList;
        foreach(KbAnim* anim, _animList){
            aList << anim->guid().toString().toUpper();
            anim->save(settings);
        }
        settings.setValue("List", aList);
    }
}

bool KbLight::needsSave() const {
    if(_needsSave)
        return true;
    foreach(KbAnim* anim, _animList){
        if(anim->needsSave())
            return true;
    }
    return false;
}
